/*
 *  MX -> ROM : convert MINIX a.out executable to TOS ROM image.
 *
 *  usage : mxrom [-h <header>] <address> [<source> [<destination>]]
 *       ( source and dest default to a.out and rom.img respectively. )
 *
 *  (C) 2000, 01 Laurent Vogel.
 */


/*
 *  principle : change text+data into hdr+jump+text+data
 *
 *  | hdr |        | text | data | ... | reloc |
 *  +-----+        +------+------+-----+-------+
 *     |               |      |            |  text + data
 *     v               v      v    <- - - -/  altered by
 *  | ROM | jmp to |      |      |            relocation.
 *  | hdr | entry  | text | data |
 *  +-----+--------+------+------+
 * 
 *  jmp to entry is a bra.s, bra or jmp to the entry point, 
 *  guaranteed not altering registers or accessing any memory.
 *  (jmp to entry is only added when the entry is not 0.)
 *
 *  The ROM header is optional. If not supplied, then the text segment
 *  should start with it, and the entry point should be 0. (a warning
 *  is output if it is not the case)
 *
 *  NOTE: minix files produced by stmx have the entry equal to 0.
 */

#define VERSION "version 1.01, March 2nd, 2001"

/*
  previous version did not allow to specify the header and the
  reloc address on the command line.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxheader.h"
#include "endian.h"


/* ROM hdr sample : (don't ask me the meaning of these!)
        bra.s begin
        .byte   1,0
        .long 0xE00020
        .long 0xE00000
        .long 0x6100
        .long 0xE00020
        .long 0
        .long 0x03060000
        .word 3
        illegal
*/

/*
unsigned char rom_hdr[] = {
	0x60, 0x1E,
	0x01, 0x00,
	0x00, 0xE0, 0x00, 0x20,
	0x00, 0xE0, 0x00, 0x00,
	0x00, 0x00, 0x61, 0x00,
	0x00, 0xE0, 0x00, 0x20,
	0x00, 0x00, 0x00, 0x00,
	0x03, 0x06, 0x00, 0x00,
	0x00, 0x03,
	0x4A, 0xFC
};
*/

static int verbose = 0;
static char *hdr_name = NULL;

/* return codes :
   1 = too short
   2 = bad header
   3 = bss not allowed in ROM!
   4 = could not access to files
   5 = bad reloc
*/

int aout2rom(long addr, char *src_name, char *dest_name)
{
  FILE *from, *reloc, *to, *hdr;
	
  struct mx_header m ;
  int len, count ;
  unsigned char buf[1024] ;
  unsigned char reloc_buf[1024];
  unsigned char jump_code[8];

  long tsize, dsize, bsize, ssize, entry, hsize ;
  long reloc_offset ;	/* offset of the reloc section inf the file */
  long reloc_origin ;   /* origin in memory */
  int err;
  long jump_code_size;

  from = fopen(src_name, "rb");
  if(from == NULL) {
    perror("");
    return 4;
  }

  /* read a.out header */
  len = fread((char *)&m, 1, sizeof(m), from);
  if(len != sizeof(m)) return 1;

  if(m.a_cpu != 0x04) return 2;
  if(m.a_flags != 0x10) return 2;
  if(GET_WORD(&m.a_magic) != 0x0301) return 2;

  tsize = GET_LONG(&m.a_text);
  dsize = GET_LONG(&m.a_data);
  bsize = GET_LONG(&m.a_bss);
  if(bsize) {
    fprintf(stderr, "Warning: BSS size 0x%lx is not null.\n", bsize);
    fprintf(stderr, "  There will be no bss in ROM file.\n");
    fprintf(stderr, "  Please use the data section instead\n");
  }
    
  ssize = GET_LONG(&m.a_syms);
  entry = GET_LONG(&m.a_no_entry);

  if(verbose) {
    fprintf(stderr, "tsize = 0x%lx, ", tsize);
    fprintf(stderr, "dsize = 0x%lx\n", dsize);
  }
  
  /* skip extra header parameters */
  if(m.a_hdrlen > sizeof(m)) {
    count = m.a_hdrlen - sizeof(m);
    len = fread(buf, 1, count, from);
    if(len != count) return 1;
  }
  
  reloc_offset = tsize + dsize + ssize + m.a_hdrlen;
  reloc = fopen(src_name, "rb");
  if(reloc == NULL) {
    perror("reloc");
    return 4;
  }
  err = fseek(reloc, reloc_offset, SEEK_SET);
  if(verbose) {
    fprintf(stderr, "reloc_offset = 0x%lx\n", reloc_offset);
  }
  if(err) {
    fprintf(stderr, "cannot seek to reloc info\n");
    return 4;
  }
  
  to = fopen(dest_name, "wb");
  if(to == NULL) {
    perror("to");
    return 4;
  }
  
  /* emit ROM hdr */
  if(hdr_name != NULL) {
    hdr = fopen(hdr_name, "rb");
    if(hdr == NULL) {
      fprintf(stderr, "could not open header file %s\n", hdr_name);
      return 4;
    }
    hsize = 0;
    do {
      count = 1024;
      count = fread(buf, 1, count, hdr);
      len = fwrite(buf, 1, count, to);
      if(len != count) {
        fprintf(stderr, "fwrite(rom_hdr, 1, %d, to) -> %d\n", count, len);
        return 4;
      }
      hsize += len;
    } while(count > 0);
    fclose(hdr);
    reloc_origin = addr + hsize;
    if(verbose) {
      fprintf(stderr, "header size 0x%lx\n", hsize);
    }
  } else {
    /* no header */
    reloc_origin = addr;
    if(entry != 0) {
      fprintf(stderr, "Warning, entry 0x%08lx is not null\n", entry);
      fprintf(stderr, "  No jump to entry point will be generated.\n");
      fprintf(stderr, "  You should check the output with a disassembler.\n");
      entry = 0;
    }
  }
  
  /* determine jump code */
  if(entry < 0 || entry & 1) {
    fprintf(stderr, "entry negative or odd\n");
    return 2;
  }
  if(entry == 0) {
    jump_code_size = 0;
  } else if(entry <= 0x7E) {
    jump_code_size = 2;
    /* bra.s _start */
    SET_WORD(jump_code, 0x6000 + (short) entry);
  } else if(entry <= 0x7FFE) {
    jump_code_size = 4;
    /* bra _start */
    SET_WORD(jump_code, 0x6000);
    SET_WORD(jump_code+2, (short) entry);
  } else {
    jump_code_size = 6;
 /*     0:  4EF9 <start>    jmp _start
  */
    SET_WORD(jump_code, 0x4EF9);
    SET_LONG(jump_code+2, entry + reloc_origin + 6);
  }
  if(jump_code_size) {
    if(verbose) {
      fprintf(stderr, "jump_code_size = 0x%08lx\n", jump_code_size);
    }
    count = jump_code_size;
    len = fwrite(jump_code, 1, count, to);
    if(len != count) return 4;
  }
  reloc_origin += jump_code_size;
  
  if(verbose) {
    fprintf(stderr, "reloc_origin = 0x%08lx\n", reloc_origin);
  }
  
  /* now, copy text and data chunk by chunk while relocating. */
  
  {
    long window_start;
    int  window_len;
    long adr ;
    unsigned char *a;
    unsigned char *p;	/* pointer to next reloc byte */
    int n;      /* number of relocs remaining in buffer */
    int c;
    
    count = fread(reloc_buf, 1, 1024, reloc);
    if(count < 4) return 1;
    
    window_start = 0;
    
    /* load first window */
    window_len = 1024;
    if(window_len > tsize + dsize) {
      window_len = tsize + dsize;
    }
    count = fread(buf, 1, window_len, from);
    if(count != window_len) return 1;
    window_len -= 4;

    adr = GET_LONG(reloc_buf);
    if(adr == 0) {
      goto no_more_reloc;
    }

    n = count - 4;
    p = reloc_buf + 4;
    
    if(verbose) {
      fprintf(stderr, "relocating at offset: ");
    }
    
    /* following code inspired from Minix 1.5 mm/exec.c */
    for (;;) {			/* once per relocation */
      /* here, relocate long at address adr 
       * (adr = 0 at beginning of text)
       * (was:) *((long *)adr) += off;
       */
       
      while(adr >= window_start + window_len) {
      	/* ship out current window */
      	count = fwrite(buf, 1, window_len, to);
        if(count != window_len)  return 4;
        window_start += window_len;
      	
      	/* copy additional end byte to the beginning of the buffer */
      	buf[0] = buf[window_len];
      	buf[1] = buf[window_len+1];
      	buf[2] = buf[window_len+2];
      	buf[3] = buf[window_len+3];
      	
      	/* load next window */
      	window_len = 1024 - 4;
      	if(window_start + window_len > tsize + dsize) {
      	  window_len = tsize + dsize - window_start;
      	}
      	if(window_len <= 0) {
      	  goto no_more_reloc;
      	}
      	count = fread(buf+4, 1, window_len, from);
      	if(count != window_len) return 1;
      }
      a = buf + (adr - window_start);
      if(verbose) {
      	fprintf(stderr, "%lx ", adr);
      }
      SET_LONG(a, GET_LONG(a) + reloc_origin);
      for (;;) {		/* once per byte */
        if (--n < 0) {
          p = reloc_buf;
          count = 1024;
          count = fread(reloc_buf, 1, count, reloc);
          if (count <= 0) {
            if(verbose) {
              fprintf(stderr, "\n");
            }
            fprintf(stderr, "no more reloc\n");
            return 1;	/* error */
          }
          n = count - 1;
        }
        c = *p++ & 0xFF;
        if (c != 1)
          break;
        adr += 254;
      }          
      if (c == 0)
        break;
      if (c & 1)
        return 5;	/* error */
      adr += c;
    }
    if(verbose) {
      fprintf(stderr, "\nend of relocation\n");
    }
      
no_more_reloc:
    window_len += 4;
    /* ship out the window, and copy the rest */
    while(window_len > 0) {
      count = fwrite(buf, 1, window_len, to);
      if(count != window_len)  return 4;
      window_start += window_len;
      window_len = 1024;
      if(window_start + window_len > tsize + dsize) {
        window_len = tsize + dsize - window_start;
      }
      if(window_len > 0) {
        count = fread(buf, 1, window_len, from);
        if(count != window_len) return 1;
      }
    }
  }
  
  fclose(to);
  fclose(from);
  fclose(reloc);
  return 0;
}

int parse_long(char *s, long *addr)
{
  long a;
  int i;
  if(*s == '0') {
    s++;
    if(*s == 'x') {
      s++;
      a = 0;
      i = 0;
      while(*s) {
        if(*s >= '0' && *s <= '9') {
          a <<= 4L;
          a += (*s - '0');
          i++;
        } else if(*s >= 'A' && *s <= 'F') {
          a <<= 4L;
          a += (*s - 'A' + 10);
          i++;
        } else if(*s >= 'a' && *s <= 'f') {
          a <<= 4L;
          a += (*s - 'a' + 10);
          i++;
        } else {
          return 0;
        }
        if(i >= 9) return 0;
        s++;
      }
      if(i == 0) return 0;
    } else {
      a = 0;
      i = 0;
      while(*s) {
        if(*s >= '0' && *s <= '7') {
          a <<= 3L;
          a += (*s - '0');
          i++;
        } else {
          return 0;
        }
        if(i >= 12) return 0;
        s++;
      }
    } 
  } else {
    a = 0;
    i = 0;
    while(*s) {
      if(*s >= '0' && *s <= '9') {
        a *= 10L;
        a += (*s - '0');
        i++;
      } else {
        return 0;
      }
      if(i >= 11) return 0;
      s++;
    }
    if(i == 0) return 0;
  } 
  *addr = a;
  return 1; 
}

int main(int argc, char **argv)
{
  char *src_name ;
  char *dest_name ;
  char *progname;
  long addr;
  int err;

  verbose = 0;
  progname = argv[0];

  while(argc > 1) {
    if(!strcmp("-h", argv[1]) || !strcmp("--header", argv[1])) {
      if(argc < 2) goto usage;
      hdr_name = argv[2];
      argv+=2;
      argc-=2;
    } else if(!strcmp("--help", argv[1])) {
      /* TODO */
      goto usage;
    } else if(!strcmp("--version", argv[1])) {
      printf(VERSION);
      printf("\n");
      exit(0);
    } else if(!strcmp("-v", argv[1]) || !strcmp("--verbose", argv[1])) {
      verbose = 1;
      argv++;
      argc--;
    } else {
      break;
    }
  }
  if(argc <= 1) {
    goto usage;
  } else {
    if(!parse_long(argv[1],&addr)) {
      goto usage;
    }
    argc --;
    argv ++;
  }
  if(argc == 1) {
    src_name = "a.out" ;
    dest_name = "rom.img" ;
  } else if(argc == 2) {
    src_name = argv[1] ;
    dest_name = "rom.img" ;
  } else if(argc == 3) {
    src_name = argv[1] ;
    dest_name = argv[2] ;
  } else {
    goto usage;
  }

  err = aout2rom(addr, src_name, dest_name) ;
  switch(err) {
  case 0:
    exit(0);
  case 4:
    perror(progname);
    exit(1);
  default:
    fprintf(stderr, "error number %d\n", err);
    exit(1);
  }
  
usage:
  fprintf(stderr, "usage: %s [-h <header>] <address> [<src> [<dest>]]\n", 
    progname);
  fprintf(stderr, 
"<src> and <dest> default to a.out and rom.img respectively\n");
  exit(1);
}
