/*
 *   ST -> MX : Atari ST gemdos executable to MINIX format.
 *
 *  usage : stmx [<source> [<destination>]]
 *           ( source and dest default to a.prg and b.out respectively. )
 */


#include <stdio.h>
#include <stdlib.h>

#include "stheader.h"
#include "mxheader.h"
#include "endian.h"


long dosymbols(FILE *from, FILE *to, 
	      long tsize, long dsize, long bsize, long ssize);

/* returns size of symbol table output. */

long dosymbols(FILE *from, FILE *to, 
	      long tsize, long dsize, long bsize, long ssize)
{
  int i, nfrom, nto;
  long len;
  unsigned short type;
  long value;
  struct nlist b;
  struct tos_symbol a;

  dsize += tsize;   /* data between tsize and dsize */
  bsize += dsize;   /* bss  between dsize and bsize */

  /* initialize unused fields */
  b.n_numaux = 0;
  SET_WORD(&b.n_type, 0);

  nfrom = ssize / sizeof(struct tos_symbol);

  nto = 0;
  while(nfrom > 0) {
    nfrom --;
    fread(&a, 1, sizeof(struct tos_symbol), from);
    type = GET_WORD(&a.sym_type);
    switch(type) {
    case 0xa248u:
      /* longer symbols are silently truncated. */
      /* FALLTHRU */
    case 0xa200u:
      /* plain symbol */
      for(i = 0 ; i < 8 ; i++) {
	b.n_name[i] = a.sym_name[i];
      }
      value = GET_LONG(&a.sym_value);
      if((value >= 0) && (value < tsize)) {
	b.n_sclass = N_TEXT;
      } else if((value >= tsize) && (value < dsize)) {
	b.n_sclass = N_DATA;
      } else if((value >= dsize) && (value < bsize)) {
	b.n_sclass = N_BSS;
      } else {
	/* strange symbol, not in range */
	fprintf(stderr, "symbol '%8s' out of range, value = %08lx\n", 
		b.n_name, value);
	b.n_sclass = N_ABS;
      }
      SET_LONG(&b.n_value, value);
      fwrite(&b, 1, sizeof(struct nlist), to);
      nto ++;
    case 0x0000u:
      /* considered to be a continuation of a previous 0xa248 longer symbol */
      break;
    default:
      fprintf(stderr, "symbol '%8s' of unknown type %04x\n", b.n_name, type);
      break;
    }
  }
  len = nto;
  len *= sizeof(struct nlist);
  return len;
}

/* return codes :
   1 = too short
   2 = bad header
*/

int sttomx(FILE *from, FILE *to)
{
  struct tos_header t ;
  struct mx_header m ;
  size_t len, tmplen ;
  char buf[1024] ;

  long tsize, dsize, bsize, ssize, stack, ssize2 ;
  int reloc_flag ;

  len = fread((char *)&t, 1, sizeof(t), from);

  if(len != sizeof(t)) return 1;

  if(t.prg_magic[0] != 0x60) return 2;
  if(t.prg_magic[1] != 0x1a) return 2;

  tsize = GET_LONG(&t.prg_tsize);
  dsize = GET_LONG(&t.prg_dsize);
  bsize = GET_LONG(&t.prg_bsize);
  ssize = GET_LONG(&t.prg_ssize);
  reloc_flag = ! (t.prg_abs[0] | t.prg_abs[1]);

  /* now, emit mx header */

  m.a_cpu = 0x04;
  m.a_flags = 0x10 ;
  SET_WORD(&m.a_magic, 0x0301);
  SET_WORD(&m.a_version, 0);
  m.a_unused = 0;
  m.a_hdrlen = 0x20;

  SET_LONG(&m.a_text, tsize);
  SET_LONG(&m.a_data, dsize);
  SET_LONG(&m.a_bss, bsize);
  SET_LONG(&m.a_no_entry, 0L);

  stack = 0x00010000L - (dsize + bsize);
  while(stack < 0L)
    stack += 0x00010000L ;
  fprintf(stderr, "%ld bytes assigned to stack+malloc area\n", stack);
  
  SET_LONG(&m.a_total, stack+tsize);

  /* by default estimate a size. If this was overestimated we will patch
     back the header at he end of the routine .  */
  ssize2 = ((ssize / sizeof(struct tos_symbol)) * sizeof(struct nlist)) ;
  SET_LONG(&m.a_syms, ssize2);

  fwrite((char *)&m, 1, sizeof(m), to);

  /* now, output text and data segts */

  len = tsize + dsize ;
  while(len > 0) {
    tmplen = 1024;
    if(tmplen > len)
      tmplen = len;
    tmplen = fread(buf, 1, tmplen, from);
    tmplen = fwrite(buf, 1, tmplen, to);
    len -= tmplen;
  }

  /* convert symb table if any */

  if(ssize) {
    ssize = dosymbols(from, to, tsize, dsize, bsize, ssize);
  }

  /* finally, output reloc info */

  do {
    len = 1024;
    len = fread(buf, 1, len, from);
    len = fwrite(buf, 1, len, to);
  } 
  while(len != 0);

  if(ssize != ssize2) {
    int err;
    /* patch back the header */
    err = fseek( to, 0L, SEEK_SET);
    if(err) {
      fprintf(stderr, "cannot patch back the header.\n");
      return 1;
    }
    SET_LONG(&m.a_syms, ssize);
    fwrite((char *)&m, 1, sizeof(m), to);
  }    

  return 0;
}

int main(int argc, char **argv)
{
  FILE *from, *to;
  char *src_name ;
  char *dest_name ;

  if(argc == 1) {
    src_name = "a.prg" ;
    dest_name = "b.out" ;
  } else if(argc == 2) {
    src_name = argv[1] ;
    dest_name = "b.out" ;
  } else if(argc == 3) {
    src_name = argv[1] ;
    dest_name = argv[2] ;
  } else {
    goto usage;
  }

  from = fopen(src_name, "rb");
  to = fopen(dest_name, "wb");

  return sttomx(from, to) ;
usage:
  fprintf(stderr, "usage: %s [<src> [<dest>]]\n", argv[0]);
  fprintf(stderr, "<src> and <dest> default to a.prg and b.out respectively\n");
  exit(1);
}
